
class Residue():
    """
    Residue class is used for plotting a bounding box around the specified residue.
    """
    def __init__(self, id=None, chain=0, padding=0):
        """
        Initialize class. Can be empty, but better initialized with residue id and chain id.

        Chain specifies which chain the residue is on and id is the id of the residue on the given chain.
        
        Parameters:
            id: int, optional
                Residue id. Count starting at 0. Default: None.
            chain: int, optional
                Chain id. Count starting at 0. Specify which chain the residue is on. Default 0 (in case of single chain). Default: 0.
            padding: int, optional
                Optional padding to bounding box. Default: 0.
        """
        if id == None:
            self._id_list = []
            self._chain_list = []
        else:
            self._id_list = [id]
            self._chain_list = [chain]
        self._pad = padding
        
    def add_residue(self, id, chain=0):
        """
        Add a new residue to the internal list of residues.
        
        Parameters:
            id: int, optional
                Residue id. Count starting at 0. Default: None.
            chain: int, optional
                Chain id. Count starting at 0. Specify which chain the residue is on. Default 0 (in case of single chain). Default: 0.
        """
        self._id_list.append(id)
        self._chain_list.append(chain)
    
    def get_res_info(self):
        """
        Returns all internal information of class
        
        Returns: 
            list:
                list of the current residues
            list:
                list of chain ID's corresponding to residues
            int:
                padding for bounding box
        """
        return self._id_list, self._chain_list, self._pad
    
    def remove_residue(self, id, chain=0):
        """
        Remove speciefied residue from internal list.

        Parameters:
            id: int, optional
                Residue id. Count starting at 0. Default: None.
            chain: int, optional
                Chain id. Count starting at 0. Specify which chain the residue is on. Default 0 (in case of single chain). Default: 0.
        """
        
        
        curr_idx = self._id_list.index(id)

        len_ = len(self._id_list) 
        while(self._chain_list[curr_idx] != chain and curr_idx < len_):
            # if first instance of residue id does not correspond to the specified chain then search the rest of the list
            curr_idx = curr_idx + 1 + self._id_list[curr_idx + 1:].index(id)
        
        if curr_idx < len_:
            del self._id_list[curr_idx]
            del self._chain_list[curr_idx]
        else: 
            print(f"No such residue found: id {id}, chain: {chain}")
    



"""
https://github.com/bunnech/holoprot/blob/main/holoprot/utils/surface.py is the base for this file. Modifications were made.

Utilities for preparing and computing features on molecular surfaces.
"""
import os
import numpy as np
from numpy.core.numeric import full
from numpy.matlib import repmat
from Bio.PDB import PDBParser, Selection
from subprocess import Popen, PIPE
from typing import Tuple, List, Dict
import trimesh 

from provis.utils import RADII, POLAR_HYDROGENS

eps = 1e-6
Surface = Tuple[np.ndarray, np.ndarray, np.ndarray, List[str], Dict[str, str]]


def read_msms(
        file_root: str) -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[str]]:
    """Read surface constituents from output files generated using MSMS.

        file_root: str
            Root name for loading .face and .vert files (produced by MSMS). Default location is data/tmp/{pdb_id}s
   
    Parameters:
        numpy.ndarray:
            vertices
        numpy.ndarray:
            faces
        numpy.ndarray:
            vertex normals
        list:
            list of res_id's from output_pdb_as_xyzrn()
    """
    vertfile = open(file_root + ".vert")
    meshdata = (vertfile.read().rstrip()).split("\n")
    vertfile.close()

    # Read number of vertices.
    count = {}
    header = meshdata[2].split()
    count["vertices"] = int(header[0])
    ## Data Structures
    vertices = np.zeros((count["vertices"], 3))
    normalv = np.zeros((count["vertices"], 3))
    atom_id = [""] * count["vertices"]
    res_id = [""] * count["vertices"]
    for i in range(3, len(meshdata)):
        fields = meshdata[i].split()
        vi = i - 3
        vertices[vi][0] = float(fields[0])
        vertices[vi][1] = float(fields[1])
        vertices[vi][2] = float(fields[2])
        normalv[vi][0] = float(fields[3])
        normalv[vi][1] = float(fields[4])
        normalv[vi][2] = float(fields[5])
        atom_id[vi] = fields[7]
        res_id[vi] = fields[9]
        count["vertices"] -= 1

    # Read faces.
    facefile = open(file_root + ".face")
    meshdata = (facefile.read().rstrip()).split("\n")
    facefile.close()

    # Read number of faces.
    header = meshdata[2].split()
    count["faces"] = int(header[0])
    faces = np.zeros((count["faces"], 3), dtype=int)
    normalf = np.zeros((count["faces"], 3))

    for i in range(3, len(meshdata)):
        fi = i - 3
        fields = meshdata[i].split()
        faces[fi][0] = int(fields[0]) - 1
        faces[fi][1] = int(fields[1]) - 1
        faces[fi][2] = int(fields[2]) - 1
        count["faces"] -= 1

    assert count["vertices"] == 0
    assert count["faces"] == 0

    return vertices, faces, normalv, res_id


def output_pdb_as_xyzrn(pdb_file: str, xyzrn_file: str) -> None:
    """
    Converts a .pdb file to a .xyzrn file.

    Parameters:
        pdb_file: str
            path to PDB File to convert (with extension)
        xyzrn_file: str
            path to the xyzrn File (with extension)
    """
    # get absolute path for xyzrn_file, needed for open
    xyzrn_file = os.path.join(os.getcwd(), xyzrn_file)

    filename = pdb_file.split("/")[-1]
    pdb_id = filename.split(".")[0]

    parser = PDBParser()
    struct = parser.get_structure(id=pdb_id, file=pdb_file)
    outfile = open(xyzrn_file, "w")

    i = 0
    for model in struct:
        for chain in model:
            for residue in chain:
                if residue.get_resname() == "HOH":
                    continue
                for atom in residue:
                    name = atom.get_name()
                    residue = atom.get_parent()
                    # Ignore hetatms.
                    if residue.get_id()[0] != " ":
                        continue
                    resname = residue.get_resname()
                    reskey = residue.get_id()[1]
                    chain = residue.get_parent().get_id()
                    atomtype = name[0]

                    color = "Green"
                    coords = None
                    if atomtype in RADII and resname in POLAR_HYDROGENS:
                        if atomtype == "O":
                            color = "Red"
                        if atomtype == "N":
                            color = "Blue"
                        if atomtype == "H":
                            if name in POLAR_HYDROGENS[resname]:
                                color = "Blue"  # Polar hydrogens
                        coords = "{:.06f} {:.06f} {:.06f}".format(atom.get_coord()[0],
                                                                  atom.get_coord()[1],
                                                                  atom.get_coord()[2])
                        insertion = "x"
                        if residue.get_id()[2] != " ":
                            insertion = residue.get_id()[2]
                        # this is the weird ID; the 9th column of the xyzrn file
                        full_id = "{}_{:d}_{}_{}_{}_{}".format(
                            chain,
                            residue.get_id()[1], insertion, resname, name, color)
                    if coords is not None:
                        outfile.write(coords + " " + RADII[atomtype] + " 1 " + str(i) + " " + full_id +  
                                      "\n")
        i += 1


def get_surface(out_path: str, density: float):
    """
    Wrapper function that reads in the output from the MSMS executable to build the protein surface.

        out_path: str
            path to output (output path from namechecker) directory. Usually data/tmp
        density: bool
            Need to pass same density as used by the MSMS binary, as the face and vert files have the density included in their names. The variable is needed for loading these files.
        
    Returns: 
        numpy.ndarray:
            vertices
        numpy.ndarray:
            faces
        numpy.ndarray:
            vertex normals
        list:
            list of res_id's from output_pdb_as_xyzrn()
        dict:
            dictionary: residues as keys, areas as values
    """
    file_base = os.path.abspath(out_path)
    file_base = f"{file_base}_out_{int(density * 10)}"

    # file_base = pdb_path.split(".")[0]
    # file_base = f"{pdb_path}_out_{int(density * 10)}"

    vertices, faces, normals, names = read_msms(file_base)
    areas = {}
    ses_file = open(file_base + ".area")
    next(ses_file)  # ignore header line
    for line in ses_file:
        fields = line.split()
        areas[fields[3]] = fields[1]

    return vertices, faces, normals, names, areas


def compute_normal(vertices: np.ndarray, faces: np.ndarray) -> np.ndarray:
    """
    Compute normals for the vertices and faces

    Parameters:
        vertices: np.ndarray 
            Vertices of the mesh
        faces: np.ndarray 
            Faces of the mesh
    
    Returns: 
        np.ndarray:
            Normals of the mesh
    """
    vertex = vertices.T
    face = faces.T
    nface = np.size(face, 1)
    nvert = np.size(vertex, 1)
    normal = np.zeros((3, nvert))
    # unit normals to the faces
    normalf = crossp(
        vertex[:, face[1, :]] - vertex[:, face[0, :]],
        vertex[:, face[2, :]] - vertex[:, face[0, :]],
    )
    sum_squares = np.sum(normalf**2, 0)
    d = np.sqrt(sum_squares)
    d[d < eps] = 1
    normalf = normalf / repmat(d, 3, 1)
    # unit normal to the vertex
    normal = np.zeros((3, nvert))
    for i in np.arange(0, nface):
        f = face[:, i]
        for j in np.arange(3):
            normal[:, f[j]] = normal[:, f[j]] + normalf[:, i]

    # normalize
    d = np.sqrt(np.sum(normal**2, 0))
    d[d < eps] = 1
    normal = normal / repmat(d, 3, 1)
    # enforce that the normal are outward
    vertex_means = np.mean(vertex, 0)
    v = vertex - repmat(vertex_means, 3, 1)
    s = np.sum(np.multiply(v, normal), 1)
    if np.sum(s > 0) < np.sum(s < 0):
        # flip
        normal = -normal
        normalf = -normalf
    return normal.T


def crossp(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Creates the cross product of two numpy arrays

    Parameters:
        x: np.ndarray
            Array 1
        y: np.ndarray
            Array 2

    Returns: 
        np.ndarray:
            (Array 1) x (Array 2)
    """
    z = np.zeros((x.shape))
    z[0, :] = np.multiply(x[1, :], y[2, :]) - np.multiply(x[2, :], y[1, :])
    z[1, :] = np.multiply(x[2, :], y[0, :]) - np.multiply(x[0, :], y[2, :])
    z[2, :] = np.multiply(x[0, :], y[1, :]) - np.multiply(x[1, :], y[0, :])
    return z

def prepare_trimesh(vertices: np.ndarray,
                 faces: np.ndarray,
                 normals: np.ndarray = None,
                 resolution: float = 1.0,
                 apply_fixes: bool = False):
    """
    Prepare the mesh surface given vertices and faces. Optionally, compute
    normals and apply fixes to mesh.

    Parameters:
        vertices: np.ndarray
            Surface vertices
        faces: np.ndarray
            Triangular faces on the mesh
        normals: np.ndarray
            Normals for each vertex
        apply_fixes: bool
            Optional application of fixes to mesh. Check fix_mesh for details on fixes. Default: False,

    Returns: 
        trimesh.Trimesh:
            Mesh
    """
    tmp_mesh = trimesh.Trimesh(vertices, faces)
    
    if apply_fixes:
        tmp_mesh.process(validate=1)

    if apply_fixes or normals is None:
        normals = compute_normal(tmp_mesh.vertices, tmp_mesh.faces)

    mesh = trimesh.Trimesh(tmp_mesh.vertices, tmp_mesh.faces, vertex_normals=normals)
    return mesh



def fix_trimesh(mesh, resolution: float = 1.0):
    """
    Applies a predefined set of fixes to the mesh, and converts it to a
    specified resolution. These fixes include removing duplicated vertices wihin
    a certain threshold, removing degenerate triangles, splitting longer edges to
    a given target length, and collapsing shorter edges.

    Parameters:
        mesh: trimesh.Trimesh
            Mesh
        resolution: float
            Maximum size of edge in the mesh

    Returns: 
        trimesh.Trimesh:
            mesh with all fixes applied
    """
    target_len = resolution
    mesh, _ = trimesh.remove_duplicated_vertices(mesh, 0.001)

    count = 0
    mesh, __ = trimesh.remove_degenerated_triangles(mesh, 100)
    mesh, __ = trimesh.split_long_edges(mesh, target_len)
    num_vertices = mesh.num_vertices
    while True:
        mesh, __ = trimesh.collapse_short_edges(mesh, 1e-6)
        mesh, __ = trimesh.collapse_short_edges(
            mesh, target_len, preserve_feature=True)
        mesh, __ = trimesh.remove_obtuse_triangles(mesh, 150.0, 100)
        if mesh.num_vertices == num_vertices:
            break

        num_vertices = mesh.num_vertices
        #print("#v: {}".format(num_vertices));
        count += 1
        if count > 10: break

    mesh = trimesh.resolve_self_intersection(mesh)
    mesh, __ = trimesh.remove_duplicated_faces(mesh)
    mesh = trimesh.compute_outer_hull(mesh)
    mesh, __ = trimesh.remove_duplicated_faces(mesh)
    mesh, __ = trimesh.remove_obtuse_triangles(mesh, 179.0, 5)
    mesh, __ = trimesh.remove_isolated_vertices(mesh)
    mesh, _ = trimesh.remove_duplicated_vertices(mesh, 0.001)

    return mesh

def find_nearest_atom(coords, res_id, new_verts):
    j = len(res_id)
    el = res_id[j-1]
    full_res_id = []
    
    from scipy import spatial
    #airports = new_verts
    tree = spatial.KDTree(coords)
    
    for i in range(len(new_verts)):
        dist, loc = tree.query(new_verts[i])
        lenr = len(res_id)
        full_res_id.append(res_id[loc])
        
    return full_res_id



from os import fchdir
from Bio.PDB import *
import pyvista as pv
import numpy as np
from os.path import exists
import pandas as pd
from biopandas.mol2 import PandasMol2
import trimesh
from biopandas.mol2 import split_multimol2
import copy

from provis.utils.atminfo import *
from provis.src.processing.name_checker import NameChecker
from provis.src.processing.file_converter import FileConverter


class DataHandler:
    """
    The 'brain' of provis, when it comes to handling atomic positions. 
    
    This class loads information from a variety of files and creates meshes to be plotted. 
    Upper level classes - eg. Structure - use DataHandler objects to create the meshes.
    
    The DataHandler class loads atom-positional information from a .pdb file and from this information computes the necessairy molecular structure mesh.
    It also loads pre-defined dictionaries from atminfo.py, that encode the size, coloring and mass of a given atom or residue.
    The member functions range from loading the atoms from the .pdb file and storing them by type, to creating the meshes from this information, as well as calculating bonds or the backbone.
    """
    def __init__(self, nc, fc=None):
        """
        Load structure form pdb file (by parsing file) and save Biopython structure representation of the protein.
        It also loads, above mentioned, pre-defined dictionaries from atminfo.py, that encode the size, coloring and mass of a given atom or residue.
        
        It also calculates the center of the atomic positions (as they appear in the pdb file). This will be used to re-center the atoms.

        Parameters:
            nc: NameChecker
                Instance of a NameChecker class. Used to pass the pdb file name and paths.
            fc: FileConverter, optional
                Instance of a FileConverter class. Needed if temporary files have not been created before instantiating this class. Default: None.
        """
        self._path, self._out_path, self._base_path, mesh_path = nc.return_all()
        if not fc:
            self._fc = FileConverter(nc)
        else:
            self._fc = fc
        
        self._fc.pdb_to_xyzrn(self._path, self._out_path)
        
        parser = PDBParser()
        file_name = self._path + ".pdb"
        self._structure = parser.get_structure(self._path, file_name)

        self._res_size_dict, self._res_color_dict = import_res_size_info()
        self._atoms_size_dict, self._atoms_color_dict, self._vw_dict = import_atm_size_info(True)
        
        #calculate the center and bounds of the atom cloud
        self._cam_pos = [0, 0, 0]
        self._max_coords = [0, 0, 0]
        self._centroid = [0, 0, 0]
        num_atoms = 0
        for model in self._structure:
            for chain in model:
                for residue in chain:
                    if residue.get_resname() == "HOH":
                        continue
                    for atom in residue:
                        num_atoms += 1
            break

        i = 0
        s = int(num_atoms / 3)
        atoms = []
        for model in self._structure:
            for chain in model:
                for residue in chain:
                    if residue.get_resname() == "HOH":
                        continue
                    for atom in residue:
                        atoms.append(atom.get_coord())
                        i += 1
            break
        center = np.array(atoms)
        self._centroid = center.mean(axis=0)
        self._max_coords = center.max(axis=0)
        self._cam_pos = [0, max(self._max_coords) * 3, 0]

    def get_structure(self):
        """
        Return the loaded structure object
        
        Returns: 
            structure
        """
        return self._structure

    def get_atoms(self, show_solvent=False, model_id=0):
        """
        Creates a dictionary that stores the 3D coordinates for each atom.
        The dictionary keys are the atom names. For each atom type in the given molecule the coordinates of the atoms of this type are stored in a list within the dictionary.
        
        The code in words:
        The .pdb file (loaded in __init__()) is iterated through. For each atom it is checked if the type of this atom is already in the dictionary. 
        If not then a new list is created with the coordinates of this atom and added to the dictionary with the name of the atom as the key.
        If the name of this atom is already present then the coordinates of the current atom are added to the list of coordinates of this same type of atom.
        The dictionary is returned.
        
        Parameters:
            show_solvent: bool, optional
                If True solvent molecules also added to retrun dictionary. Default: False.
            model_id: int, optional
                The dynamic model ID of the desired molecule. Count starts at 0. Leave default value for static molecules. Default: 0.
        
        Returns: 
            dict
                Dictionary of atomic coordinates by atom type.
            int
                Maximum coordinate in y axis. (Used to create default camera)
        """
        calculate = True
        for i in range(3):
            if self._centroid[i] != 0:
                calculate *= False

        if calculate:
            num_atoms = 0
            for model in self._structure:
                if model.id == model_id:
                    for chain in model:
                        for residue in chain:
                            if residue.get_resname() == "HOH" and not show_solvent:
                                continue
                            for atom in residue:
                                num_atoms += 1

            i = 0
            s = int(num_atoms / 3)
            atoms = []
            for model in self._structure:
                if model.id == model_id:
                    for chain in model:
                        for residue in chain:
                            if residue.get_resname() == "HOH" and not show_solvent:
                                continue
                            for atom in residue:
                                atoms.append(atom.get_coord())
                                i += 1
            center = np.array(atoms)
            self._centroid = center.mean(axis=0)
            self._max_coords = center.max(axis=0)
            self._cam_pos = [0, max(self._max_coords) * 3, 0]
        i = 0
        atom_data = dict()
        for model in self._structure:
            if model.id == model_id:
                for chain in model:
                    for residue in chain:
                        if residue.get_resname() == "HOH" and not show_solvent:
                            continue
                        for atom in residue:
                            type_name = atom.element
                            # If atom not in dictionary, add it as key with coords in list
                            if type_name not in atom_data:
                                coords = copy.deepcopy(atom.get_coord())
                                for j in range(3):
                                    coords[j] -= self._centroid[j]
                                atom_data[type_name] = [coords]
                            # If atom already in dictionary, append its coordinates to list
                            else:
                                coords = copy.deepcopy(atom.get_coord())
                                for j in range(3):
                                    coords[j] -= self._centroid[j]
                                atom_data[type_name].append(coords)
            i += 1

        # return the 3D positions of atoms organized by atom type
        return atom_data
        
            
    def get_atoms_IDs(self, model_id=0):
        """
        Get dictionary of atomic coordinates (same as get_atoms()) and residue IDs (format from output_pdb_as_xyzrn()) from the xyzrn file.
        Also return a list of all the atomic coordinates in a list in the same order as in the .pdb file.
        
        Used by the Surface class.
        
        The code in words:
        The .xyzrn file is loaded and is iterated through. The relevant fields are stored to temporary variables.
        For each atom it is checked if the type of this atom is already in the dictionary. 
        If not then a new list is created with the coordinates of this atom and added to the dictionary with the name of the atom as the key.
        If the name of this atom is already present then the coordinates of the current atom are added to the list of coordinates of this same type of atom.
        Regardless of the atom already being present in the dictionary the residue id and the coordinates are added to the two lists.
        The dictionary and the two lists are returned.
        
        Parameters:
            model_id: int, optional
                The dynamic model ID of the desired molecule. Count starts at 0. Leave default value for static molecules. Default: 0.
        
        Returns: 
            dict
                Dictionary of atomic coordinates by atom type.
            list
                List of unique residue IDs (format from output_pdb_as_xyzrn())
            list
                Atomic coordinates (in same order as the residue IDs)
        """
        xyzrnfile = open(self._out_path + ".xyzrn")
        meshdata = (xyzrnfile.read().rstrip()).split("\n")
        xyzrnfile.close()
       
        res_id = []
        atom_data = dict()
        atom_coords = []
        lenm = len(meshdata)
        for vi in range(lenm):
            fields = meshdata[vi].split()
            if int(fields[5]) == model_id:
                vertices = [None] * 3
                vertices[0] = float(fields[0]) - self._centroid[0]
                vertices[1] = float(fields[1]) - self._centroid[1]
                vertices[2] = float(fields[2]) - self._centroid[2]
                res_id.append(fields[6])
                res = fields[6].split("_")
                atmtype = res[4][0]
                if atmtype not in atom_data:
                    atom_data[atmtype] = [vertices]
                # If atom already in dictionary, append its coordinates to list
                else:
                    atom_data[atmtype].append(vertices)
                atom_coords.append(vertices)

        # return the 3D positions of atoms organized by atom type
        return atom_data, res_id, atom_coords

    def get_residues(self, model_id=0, show_solvent=False):
        """
        Creates a dictionary of coordinates by residues from structure object
        
        The code in words:
        The .pdb file (loaded in __init__()) is iterated through. For each residue it is checked if the type of this residue is already in the dictionary. 
        If not then a new list is created with the coordinates of this residue and added to the dictionary with the type of the residue as the key.
        If the type of this residue is already present then the coordinates of the current residue are added to the list of coordinates of this same type of residue.
        The coordinates of the residue are calculated as the arithmetic center of the coordinates.
        The dictionary is returned.
        
        Parameters:
            model_id: int, optional
                The dynamic model ID of the desired molecule. Count starts at 0. Leave default value for static molecules. Default: 0.
            show_solvent: bool, optional
                If True solvent molecules also added to retrun dictionary. Default: False.
        
        Returns: 
            dict
                Dictionary of atomic coordinates by residue type.
        """

        # load file into a python list
        residues = self._structure.get_residues()
        res_list = list(residues)

        # initialize dictionary, then fill up iteratively
        res_data = dict()
        for model in self._structure:
            if model.id == model_id:
                for chain in model:
                    for res in chain:
                        if res.get_resname() == "HOH" and not show_solvent:
                            continue
                        else:
                            type_name = res.get_resname()
                            center_coord = [0., 0., 0.]
                            c = 0.
                            for atom in res:
                                center_coord = center_coord + atom.get_coord()
                                center_coord -= self._centroid
                                c += 1.
                            center_coord = center_coord / c
                            if type_name not in res_data:
                                res_data[type_name] = [center_coord]
                            else:
                                res_data[type_name].append(center_coord)

        # return the 3D positions of atoms organized by atom type
        return res_data

    def get_residue_info(self, res, chain, option):
        """
        Calculates information about specified residue from mol2 file.
        Depending on what is specified, either the center of mass (COM) or the charge is computed.
        
        Parameters:
            res: str
                Residue number of specified residue be looked at.
            chain:                choose what property of residue you want. com for Centre Of Mass, ch for charge
                Chain number of corresponding to residue be looked at. str - options: com, ch
        
        Returns: 
            list
                List of COM coords of given (exact) residue.
        """
        # load info for given residue
        fname = self._out_path + ".mol2"#TODO: make so it works with trajectories
        if not exists(fname):
            self._fc.pdb_to_mol2(self._path, self._out_path)
        pmol = PandasMol2().read_mol2(fname)
        my = pmol.df[pmol.df['subst_id'] == res]
        
        if len(my) == 0:
            print("ERROR: No such residue: ", res)
            return 1
        
        fid = my.iloc[0]['atom_id'] -1
        curr_chain = 0
        actual = pd.DataFrame(columns = list(my))
        for i, row in my.iterrows():
            if row['atom_id'] != fid + 1:
                curr_chain += 1
            if curr_chain == chain:
                actual = actual.append(row)
            fid = row['atom_id']
                
        # if com option calculate Centre of mass
        if option.upper() == "COM":
            COM_sum = [0.0, 0.0, 0.0]
            mass_dict = import_atm_mass_info()
            a = 0
            for i in range(len(actual)):
                m = mass_dict[actual.iloc[i]['atom_type'][0]]
                COM_sum[0] += actual.iloc[i]['x'] * m
                COM_sum[1] += actual.iloc[i]['y'] * m
                COM_sum[2] += actual.iloc[i]['z'] * m
                a += m
            if a == 0:
                print("ERROR: residue empty -> center of mass could not be found")
                return 1
            COM_sum -= self._centroid * a
            COM = [i/a for i in COM_sum]
            return COM
            
        # if charge option calculate residue charge
        elif option.upper() == "CH" or option.upper() == "CHARGE":
            charge = 0.0
            for i in range(len(my)):
                charge += my.iloc[i]['charge']

            return charge

    
    def get_atom_mesh(self, atom_data, vw=0, probe=0, phi_res=10, theta_res = 10):
        """
        Create a list of Shperes and colors representing each atom for plotting. Can later be added to a mesh for plotting.
        
        The code in words:
        Iterates through the atom_data dictionary by atom type (from get_atoms()).
        It creates uniform Spheres (same size and color) in the position specified by the coordinates list for each atom type.
        Also differentiates between Van der Waals and normal radii and handles unkown atoms.
        
        Parameters:
            atom_data: dict
                Dictionary of atoms and their coordinates, by atom type.
            vw: bool, optional
                ptional - When set to True Van-der-Waals atomic radii used instead of empirical radii. Default: False.
            probe: int, optional
                size of probe (representing the solvent size) needed for surface calculation. Default: 0.
            phi_res: int, optional
                pyvista phi_resolution for Sphere objects representing atoms. Default: 10.
            theta_res: int, optional
                pyvista theta_resolution for Sphere objects representing atoms. Default: 10.
        
        Returns: 
            list
                List of pyvista Shperes representing each atom
            list
                List of colors corresponding to each atom
            list
                List of atom ID's for each atom
        """
        # these two dictionaries have to be manually created
        # also, if more/other atoms present in protein it will not work

        # create glyphs (spherical) to represent each atom
        atoms_spheres = []
        colors_list = []
        name_list = []
        rad = probe / 2.0
        for atoms_type in atom_data:

            # create a mesh with each atoms position
            mesh = pv.PolyData(np.array(atom_data[atoms_type]))
            # place a specific sphere at given position
            if not vw:
                sphere = pv.Sphere(radius=self._atoms_size_dict[atoms_type], phi_resolution=phi_res, theta_resolution=theta_res)
            else:
                sphere = pv.Sphere(radius=(self._vw_dict[atoms_type] + rad), phi_resolution=phi_res, theta_resolution=theta_res)
            glyphs = mesh.glyph(geom=sphere)
            atoms_spheres.append(glyphs)

            try:
                # color the sphere according to 'CPK' standard
                colors_list.append(self._atoms_color_dict[atoms_type])
                name_list.append(atoms_type)
            except KeyError:
                # if atom is unrecognized, color it pink
                colors_list.append('#DD77FF')
                name_list.append("Unkown")

        # return list of spheres and colors representing each atom
        return atoms_spheres, colors_list, name_list

    def get_atom_trimesh(self, atom_data, vw=False, probe=0):
        """
        Create a list of shperes and colors representing each atom for plotting in a Trimesh format. Used for feature computation in the surface_handler class.
        
        The code in words:
        Iterates through the atom_data dictionary by atom type (from get_atoms()).
        It creates uniform Spheres (same size and color) in the position specified by the coordinates list for each atom type.
        Also differentiates between Van der Waals and normal radii and handles unkown atoms.
        
        Parameters:
            atom_data: dict
                Dictionary of atoms and their coordinates, by atom type.
            vw: bool, optional
                ptional - When set to True Van-der-Waals atomic radii used instead of empirical radii. Default: False.
            probe: int, optional
                size of probe (representing the solvent size) needed for surface calculation. Default: 0.
        
        Returns: 
            list
                List of pyvista Shperes representing each atom
            list
                List of colors for each atom
            list
                List of atom ID's for each atom
        """
        # these two dictionaries have to be manually created
        # also, if more/other atoms present in protein it will not work

        # create glyphs (spherical) to represent each atom
        atoms_spheres = []
        colors_list = []
        name_list = []
        rad = probe / 2.0
        

        for atoms_type in atom_data:

            spheres = []
            for atom in atom_data[atoms_type]:
                if not vw:
                    spheres.append(trimesh.primitives.Sphere(radius=self._atoms_size_dict[atoms_type], center=atom))
                else:
                    spheres.append(trimesh.primitives.Sphere(radius=(self._vw_dict[atoms_type] + rad), center=atom))

            mesh = trimesh.util.concatenate(spheres)
            atoms_spheres.append(mesh)

            try:
                # color the sphere according to 'CPK' standard
                colors_list.append(self._atoms_color_dict[atoms_type])
                name_list.append(atoms_type)
            except KeyError:
                # if atom is unrecognized, color it pink
                colors_list.append('#DD77FF')
                name_list.append("Unkown")

        # return list of spheres and colors representing each atom
        return atoms_spheres, colors_list, name_list

    def get_residue_mesh(self, res_data, phi_res = 25, theta_res = 25):
        """
        Create a list of Shperes and colors representing each residue for plotting. Can later be added to a mesh for plotting.
        
        The code in words:
        Iterates through the res_data dictionary by atom type (from get_residues()).
        It creates uniform Spheres (same size and color) in the position specified by the coordinates list for each residue type.
        Also differentiates between Van der Waals and normal radii and handles unkown residues.
        
        Parameters:
            res_data: dict
                Dictionary of residues and their coordinates by residue type.
            phi_res: int, optional
                pyvista phi_resolution for Sphere objects representing residues. Defaul: 25.
            theta_res: int, optional
                pyvista theta_resolution for Sphere objects representing residues. Default:25.
        
        Returns: 
            list
                List of pyvista Shperes representing each residue
            list
                List of colors for each residue
            list
                List of residue names
        """

        # create glyphs (spherical) to represent each res
        res_spheres = []
        colors_list = []
        res_names = []
        for res_type in res_data:

            sphere = pv.Sphere(radius=0.0, phi_resolution=phi_res, theta_resolution=theta_res)
            try:
                sphere = pv.Sphere(radius=self._res_size_dict[res_type], phi_resolution=phi_res, theta_resolution=theta_res)
                
                #color the sphere according to 'CPK' standard
                colors_list.append(self._res_color_dict[res_type])
                res_names.append(res_type)
                
            except KeyError:
                # color unkown to light purple
                colors_list.append('#DD77FF')
                res_names.append("Unkown")

            # create a mesh with each residues position
            mesh = pv.PolyData(np.array(res_data[res_type]))

            # place a specific sphere at given position
            glyphs = mesh.glyph(geom=sphere)
            res_spheres.append(glyphs)


        # return list of spheres and colors representing each res
        return res_spheres, colors_list, res_names

    def get_bond_mesh(self, model_id=0):
        """
        Determine bonds from 3D information.
        
        The color information is as follows:
            White for all single bonds, 
            Blue for all double bonds,
            Green for all triple bonds,
            Red for all amide bonds,
            Purple for all aromatic bonds,
            Black for everything else.
              
        The code in words:
        Parse mol2 file (also works on multi model file).
        Find where the boundaries of the current molecule are in the file.
        Extract the atomic and bond information by creating DataFrame.
        From the DataFrames get the information corresponding to the current bond: create a pyvista.Line() and store the bond type.
        Return the compiled lists of Lines and bond types (colors).
        
        Parameters:
            model_id: int, optional
                The dynamic model ID of the desired molecule. Count starts at 0. Leave default value for static molecules. Default: 0.
        
        Returns: 
            list
                List of pyvista lines representing each bond.
            list
                List of colors corresponding to the lines in the above list
            list
                List of names of the bonds: single, double, triple, amide, aromatic, unkown
        """
        calculate = True
        for i in range(3):
            if self._centroid[i] != 0:
                calculate *= False

        if calculate:
            num_atoms = 0
            for model in self._structure:
                if model.id == model_id:
                    for chain in model:
                        for residue in chain:
                            if residue.get_resname() == "HOH" :
                                continue
                            for atom in residue:
                                num_atoms += 1

            i = 0
            s = int(num_atoms / 3)
            atoms = []
            for model in self._structure:
                if model.id == model_id:
                    for chain in model:
                        for residue in chain:
                            if residue.get_resname() == "HOH":
                                continue
                            for atom in residue:
                                atoms.append(atom.get_coord())
                                i += 1
            center = np.array(atoms)
            self._centroid = center.mean(axis=0)
            self._max_coords = center.max(axis=0)
            self._cam_pos = [0, max(self._max_coords) * 3, 0]

        # Start of actual bond info calculation
        fname = self._out_path + ".mol2"
        # Check if mol2 file exists. If not convert it from pdb
        file_exists = exists(fname)
        if not file_exists:
            self._fc.pdb_to_mol2(self._path, self._out_path)
            

        k = 0
        bonds = []
        col = []
        names = []
        for mol2 in split_multimol2(fname):
            # only do computation on desired model_id
            keep_molecule = k == model_id
            if keep_molecule:
                # easy to handle the mol2 content as text -> convert it from list
                str_mol2 = ''.join(mol2[1])
                
                # find start and end of the @<TRIPOS>ATOM section
                start_atom = str_mol2.find('@<TRIPOS>ATOM')
                # 13 = len('@<TRIPOS>ATOM')
                end_atom = start_atom + 13 + str_mol2[start_atom:].replace('@<TRIPOS>ATOM','').find('@')
                # figure out number of atoms
                num_atoms = int(len(str_mol2[start_atom:end_atom].replace('@<TRIPOS>ATOM\n','').replace('\n',' ').strip().split())/9)
                # create a dataframe of atoms
                df_atoms = pd.DataFrame(np.array(str_mol2[start_atom:end_atom].replace('@<TRIPOS>ATOM\n','').replace('\n',' ').strip().split()).reshape((num_atoms,9)), columns=['atom_id', 'atom_name', 'x', 'y', 'z', 'atom_type', 'subst_id', 'subst_name', 'charge'])
                
                # find start of the @<TRIPOS>BOND section (end not needed, goes to end of string)
                start_bond = str_mol2.find('@<TRIPOS>BOND')
                # figure out number of bonds
                num_bonds = int(len(str_mol2[start_bond:].replace('@<TRIPOS>BOND\n','').replace('\n',' ').strip().split())/4)
                # create a dataframe of bonds
                df_bond = pd.DataFrame(np.array(str_mol2[start_bond:].replace('@<TRIPOS>BOND\n','').replace('\n',' ').strip().split()).reshape((num_bonds,4)), columns=['bond_id', 'atom1', 'atom2', 'bond_type'])
                
                # find each the 3D coords for each two atoms in a bond
                for i in range(num_bonds):
                    my = df_atoms.iloc[int(df_bond.iloc[i]['atom1'])-1][['x', 'y', 'z']]
                    c = df_atoms.iloc[int(df_bond.iloc[i]['atom2'])-1][['x', 'y', 'z']]
                    
                    my_coord = my.values.tolist()
                    c_coord = c.values.tolist()
                    my_coord = [float(i) for i in my_coord]
                    c_coord = [float(i) for i in c_coord]
                    my_coord -= self._centroid
                    c_coord -= self._centroid
                    if df_bond.iloc[i]['bond_type'] == '1':
                        line = pv.Line(my_coord, c_coord, resolution=5)
                        col.append('w')
                        names.append('single')
                    elif df_bond.iloc[i]['bond_type'] == '2':
                        line = pv.Line(my_coord, c_coord, resolution=1)
                        col.append('b')
                        names.append('double')
                    elif df_bond.iloc[i]['bond_type'] == '3':
                        line = pv.Line(my_coord, c_coord, resolution=1)
                        col.append('g')
                        names.append('triple')
                    # amide bond in red
                    elif df_bond.iloc[i]['bond_type'] == 'am':
                        line = pv.Line(my_coord, c_coord, resolution=1)
                        col.append('r')
                        names.append('amide')
                    # aromatic bond in purple
                    elif df_bond.iloc[i]['bond_type'] == 'ar':
                        line = pv.Line(my_coord, c_coord, resolution=1)
                        col.append('m')
                        names.append('aromatic')
                    else:
                        line = pv.Line(my_coord, c_coord, resolution=1)
                        col.append('k')
                        names.append('unkown')
                    bonds.append(line)
                break
            k += 1
        
        return bonds, col, names
    
    def get_backbone_mesh(self, model_id=0):
        """
        Creates and returns a Spline object representing the backbone of the protein.
          
        The code in words:
        Iterates through the res_data dictionary by atom type (from get_residues()).
        Calculates the center of each residue and returns these points as a numpy array.
        (Later used to create a Spline.)
             
        Parameters:         
            model_id: int, optional
                The dynamic model ID of the desired molecule. Count starts at 0. Leave default value for static molecules. Default: 0.
        
        Returns: 
            pyvista.Spline
                Spline running through coordinates representing the centre of mass of each residue.
        """
        
        i = 0
        res_list = []
        for model in self._structure:
            if model.id == model_id:
                for chain in model:
                    for res in chain:
                        if res.get_resname() != "HOH":
                            com = [0, 0, 0]
                            count = 0
                            for atom in res:
                                coords = atom.get_coord()
                                com += coords
                                com -= self._centroid
                                count += 1
                            com /= count
                            res_list.append(com)
            i += 1
        
        return pv.Spline(np.array(res_list))



import pyvista as pv
from os.path import exists
from provis.src.processing.data_handler import DataHandler
from provis.src.processing.name_checker import NameChecker

class Structure:
    """
    The Structure class is used to visualize the structural information of the given molecule. One can easily plot the atoms, residues, bonds or any combination of these structures.
    """
    def __init__(self, nc, dh=None, plot_solvent=False, notebook=False, msms=False):
        """
        In this constructor no mesh information is computed, simply preparations are made. Meshes are loaded in the plotting function.
        
        A NameChecker object is required for initialization, as this is how the program finds the desired pdb file, molecule.
        If nothing else is passed the NameChecker object will be used to initialize the other internal objects of the structure class.
        
        Apart from the required NameChecker object one can also pass a DataHandler for even more control.
        
        Parameters:
            nc: NameChecker
                Instance of a NameChecker class. Used to pass the pdb file name and paths.
            dh: DataHandler, optional
                Instance of a DataHandler class. Used to retrieve atom-positional information. Default: None. If None a new DataHandler variable will be initialized with "nc".
            plot_solvent: bool, optional
                If True solvent molecules will also be plotted. Default: False.
            notebook: bool, optional 
                Needs to be set to true to work in a notebook environment. Defualts to False.
            msms: bool, optional
                Set to True when running the msms version. Only used to save image with "msms" at end of filename ({root directory}/data/img/{pdb_id}_{plot type}_msms.png). Default: False.
        """
        self._msms = msms
        self._notebook = notebook
        self._shading = not self._notebook
        self._solvent = plot_solvent
        self._path, self._out_path, self._base_path, self._mesh_path = nc.return_all()
        # create "brain" of plotting class
        if not dh:
            self._dh = DataHandler(nc)
        else:
            self._dh = dh
        self._cam_pos = [0, 0, 0]

    def manual_plot(self, box=0, res=0, outname=0, atoms=0, col_a=0, bonds=0, vw=0, residues=0, col_r=0, bb=0, camera=None):
        """
        Plot stick and point model. In this function one can pass all the desired meshes to be plotted. One can get these meshes from the DataHandler class.
        
        Parameters:
            box: bool, optional
                If True bounding box also visualized, default: 0.
            res: list, optional
                List of pyvista Shperes representing each residue, default: 0.
            outname: string, optional
                save image of plot to specified filename. Will appear in data/img directory. default: data/img/{self._out_path}_stick_point.
            atoms: list, optional
                List of pyvista Shperes representing each atom, default: 0.
            col_a: list, optional
                List of colors for each atom, default: 0.
            bonds: list, optional
                List of pyvista Lines representing each bond, default: 0.
            vw: bool, optional
                If True styling for Van-der-Waals plotting set. Vw atomic objects still have to be passed under 'atoms' variable.
            col_r: list, optional
                List of colors for each residue, default: 0.
            res: Residue, optional
                Specified residues will be plotted with a bounding box around them.
            bb: bool, optional
                List of coordinates describing the back-bone of the protein, default: 0.
            camera: pyvista.Camera, optional
                Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to [0, 4 * "max distance from the center", 0] (see: https://pro-vis.readthedocs.io/en/latest/tutorial.html for more detail). Default: None.
        
        Returns: 
            Pyvista.Plotter window
                Window with interactive plot
        """
        
        pl = pv.Plotter(notebook=self._notebook)
        pl.background_color = 'grey'
        pl.enable_3_lights()

        # adding the spheres (by atom type) one at a time
        opacity = 1 - vw*0.4
        style = 'surface'
        if atoms != 0:
            if vw:
                style='wireframe'
            for j, mesh in enumerate(atoms):
                pl.add_mesh(mesh, color=col_a[j], opacity=opacity, smooth_shading=self._shading, style=style)

        # adding the bonds one at a time
        if bonds != 0:
            for b in bonds:
                pl.add_mesh(b, color='w', render_lines_as_tubes=True, line_width=5)
        # adding the spheres (by residue) one at a time
        # only executes if residue information provided
        if residues != 0:
            for k, mesh in enumerate(residues):
                pl.add_mesh(mesh, color=col_r[k], opacity=0.2)
        
        # if specified add bounding box
        if box:
            pl.add_bounding_box(color='white', corner_factor=0.5, line_width=1)
        if res:
            res_list, chain_list, pad = res.get_res_info()
            for i, r in enumerate(res_list):
                chain = chain_list[i]
                residues = self._dh.get_structure().get_residues()
                residues_list = list(residues)
                res_name = residues_list[r + 1].get_resname()
                d = (self._dh._res_size_dict[res_name] + pad) * 2
                x, y, z = d,d,d
                pl.add_mesh(pv.Cube(center=self._dh.get_residue_info(r, chain,'com'), x_length=x, y_length=y, z_length=z), style='wireframe', show_edges=1, line_width=5, color='r')
        
        
        if bb:
            pl.add_mesh(bb, render_lines_as_tubes=True, smooth_shading=self._shading, line_width=10)
        
        # save a screenshot
        if not outname:
            new_name = self._out_path.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_stick_point.png'
        
        if camera: 
            pl.camera = camera
        else:
            self._cam_pos = self._sh._dh._cam_pos
            pl.camera.position = self._cam_pos
        pl.show(screenshot=outname, title='Provis')


    def plot(self, box=0, res=None, outname=0, atoms=0, bonds=0, vw=0, residues=0, bb=0, title=None, camera=None, model_id=0, dynamic=False):
        """
        This member function is called by all the others. Using this function you can plot any combination of the results gotten from the specialized member functions. For example you could plot the atoms and the backbone of the protein in the same plot.
        
        All information to be plotted is already computed. This function simply dictates what is to be plotted.
        
        Parameters:
            box: bool, optional
                ptional - If True bounding box also visualized, default: 0.
            res: Residue, optional
                Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
            outname: string, optional
                Save image of plot to specified filename. Will appear in data/img directory. Defaults to {root directory}/data/img/{pdb_id}_{model_id}_stick_point.png. If Structure class was initialized with msms=True then output will have "_msms.png" as the ending.
            atoms: bool, optional
                Plot atoms, default: 0.
            bonds: int, optional
                ptional - Plot bond. If zero or undefined then it does not plot the bonds, if 1 it plots all bonds uniformly, if 2 it plots colorful bonds (see data_handler). Default: 0.
            vw: bool, optional
                Plot Wan-der-Waals radii instead of atomic radii.
            residues: bool, optional
                ptional - Plot residue, default: 0.
            bb: bool, optional
                If True backbone of protein is plotted. Default: False.
            title: str, optional
                Title of the plot window. Defaults to None.
            camera: pyvista.Camera, optional
                Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to [0, 4 * "max distance from the center", 0] (see: https://pro-vis.readthedocs.io/en/latest/tutorial.html for more detail). Default: None.
            model_id: int, optional
                The dynamic model ID of the desired molecule. Count starts at 0. Leave default value for static molecules. Default: 0.
            dynamic: bool, optional
                Set to True if you are plotting a dynamic model. Default: False.
        
        Returns: 
            Pyvista.Plotter window
                Window with interactive plot
        """
        atom_data = self._dh.get_atoms(show_solvent=self._solvent, model_id=model_id) # second arg: 1 = show_solvent
        self._cam_pos = self._dh._cam_pos
        pl = pv.Plotter(notebook=self._notebook)
        print("Structure plotter created for model id: ", model_id)
        pl.background_color = 'grey'
        pl.enable_3_lights()
            
        if camera: 
            pl.camera = camera
            print("Camera added...")
        else:
            pl.camera_position = self._cam_pos
              
        # adding the spheres (by atom type) one at a time
        opacity = 1 - vw*0.4
        style = 'surface'
        if atoms:
            if vw:
                _atoms_vw, _col_vw, _ = self._dh.get_atom_mesh(atom_data, vw=1)
                style='wireframe'
                for j, mesh in enumerate(_atoms_vw):
                    pl.add_mesh(mesh, color=_col_vw[j], opacity=opacity, smooth_shading=self._shading, style=style)
                print("Van-der-Waals atoms added...")
            else:
                ## return list of spheres (meshes) and colors for the spheres
                _atoms, _col_a, _ = self._dh.get_atom_mesh(atom_data, vw=0) # second arg: 1 = showvw spheres instead of "normal" radius
                for j, mesh in enumerate(_atoms):
                    pl.add_mesh(mesh, color=_col_a[j], opacity=opacity, smooth_shading=self._shading, style=style)
                print("Atoms added...")

        if bonds:
            # adding the bonds one at a time
            bond_mesh_name = self._mesh_path + "_" + str(model_id) + "_bonds"
            if bonds == 1:
                bond_mesh_name += ".vtk"
                if exists(bond_mesh_name):
                    mesh = pv.PolyData(bond_mesh_name)
                else:
                    print("Calculating Bond mesh")
                    ## return list of lines (meshes)
                    _bonds, _bond_col, _ = self._dh.get_bond_mesh()
                    mesh = pv.PolyData()
                    for b in _bonds:
                        mesh += b
                    mesh.save(bond_mesh_name)
                pl.add_mesh(mesh, color="w", line_width=5, render_lines_as_tubes=True)
            if bonds == 2:
                _bonds, _bond_col, _ = self._dh.get_bond_mesh()
                for b, c in zip(_bonds, _bond_col):
                    pl.add_mesh(b, color=c, line_width=5, render_lines_as_tubes=True)
                # colors = []
                # mesh = pv.PolyData()
                # bond_col_name = bond_mesh_name + "_col"
                # bond_mesh_name += "_col.stl"         
                # if exists(bond_mesh_name) and exists(bond_col_name):
                #     mesh = pv.PolyData(bond_mesh_name)
                #     with open(bond_col_name, 'rb') as f:
                #         colors = pickle.load(f)
                # else:
                #     # # return list of lines (meshes)
                #     # _bonds, _bond_col, _ = self._dh.get_bond_mesh()
                #     # mesh = pv.PolyData()

                #     # for b, c in zip(_bonds, _bond_col):
                #     #     pl.add_mesh(b, color=c, line_width=5, render_lines_as_tubes=True)
                    
                #     # return list of lines (meshes)
                    
                #     _bonds, _bond_col, _bond_names = self._dh.get_bond_mesh()
                #     print(len(_bond_names))
                #     mesh = pv.PolyData()
                #     colors = []
                #     for j, b in enumerate(_bonds):
                #         mesh += b
                #         colors += [_bond_names[j]] * len(b.points)
                #     import pickle
                #     print(len(mesh.points))
                #     with open(bond_col_name, 'wb') as f:
                #         pickle.dump(colors, f)
                # pl.add_mesh(b, scalars=colors, line_width=5, render_lines_as_tubes=True)
            print("Bonds added...")
                
                
        # adding the spheres (by residue) one at a time
        # only executes if residue information provided
        if residues:
            ## return list of spheres (meshes) and colors for the spheres
            res_data = self._dh.get_residues()
            _residues, _col_r, _ = self._dh.get_residue_mesh(res_data)
            
            for k, mesh in enumerate(_residues):
                pl.add_mesh(mesh, color=_col_r[k], opacity=0.2)
            print("Residues added...")

        if bb:
            bb_mesh = self._dh.get_backbone_mesh()
            pl.add_mesh(bb_mesh, render_lines_as_tubes=True, line_width=10)
            print("Back-bone added...")
             
        if res:
            res_list, chain_list, pad = res.get_res_info()
            for i, r in enumerate(res_list):
                chain = chain_list[i]
                residues = self._dh.get_structure().get_residues()
                residues_list = list(residues)
                res_name = residues_list[r + 1].get_resname()
                d = (self._dh._res_size_dict[res_name] + pad) * 2
                x, y, z = d,d,d
                
                center = self._dh.get_residue_info(r, chain,'com')
                # if residue not found 1 is returned. Otherwise the coordinates
                if center != 1:
                    pl.add_mesh(pv.Cube(center=center, x_length=x, y_length=y, z_length=z), style='wireframe', show_edges=1, line_width=5, smooth_shading=self._shading, color='r')            
            print("Residues marked...")
       
        # if specified add bounding box 
        if box:
            pl.add_bounding_box(color='white', corner_factor=0.5, line_width=1)
            print("Bounding box added...")
            
        # save a screenshot
        if not outname or outname[0] == '_':
            ending = ".png"
            if self._msms:
                ending = "_msms" + ending
                
            new_name = self._out_path.split('/')
            new_name = new_name[-1].split('.')[0]
            
            ident = '_stick_point'
            if outname:
                ident = outname
            outname = self._base_path + 'data/img/' + new_name + "_" + str(model_id) + ident + ending 
          
        print("Showing plot")  
        pl.show(screenshot=outname, title=title)


    def plot_stick_point(self, box=0, res=None, outname=0, camera=None):
        """
        Plot stick and point model of the protein. Atoms are spheres, bonds are tubes.
        
        Parameters:
            box: bool, optional
                ptional - If True bounding box also visualized, default: 0.
            res: Residue, optional
                Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
            outname: string, optional
                Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_stick_point.png.
            camera: pyvista.Camera, optional
                Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to [0, 4 * "max distance from the center", 0] (see: https://pro-vis.readthedocs.io/en/latest/tutorial.html for more detail). Default: None.
        
        Returns: 
            Pyvista.Plotter window
                Window with interactive plot.
        """

        self.plot(atoms=1, box=box, vw=0, bonds=1, residues=0, res=res, outname=outname, title="Stick Point", camera=camera)
        
    def plot_atoms(self, box=0, res=0, outname=None, camera=None):
        """
        Plot the atoms as spheres. Each atom has a radius proportianal to its calculated atomic radius.
        
        Consult https://en.wikipedia.org/wiki/CPK_coloring for the coloring. 
        
        Parameters:
            box: bool, optional
                ptional - If True bounding box also visualized, default: 0.
            res: Residue, optional
                Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
            outname: string, optional
                Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_atoms.png.
            camera: pyvista.Camera, optional
                Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to [0, 4 * "max distance from the center", 0] (see: https://pro-vis.readthedocs.io/en/latest/tutorial.html for more detail). Default: None.
        
        Returns: 
            Pyvista.Plotter window
                Window with interactive plot.
        """
        if not outname:
            outname = '_atoms'
        self.plot(atoms=1, box=box, vw=0, bonds=0, residues=0, res=res, outname=outname, title="Atoms", camera=camera)
        
    def plot_residues(self, box=0, res=0, outname=0, camera=None):
        """
        Plot the residues as Spheres. Each sphere is the approximate size of the radius of the given residue. This plot should only be used to get a general feel for the layout of the protein.
        
        For coloring information please visit: http://acces.ens-lyon.fr/biotic/rastop/help/colour.htm
        
        Parameters:
            box: bool, optional
                ptional - If True bounding box also visualized, default: 0.
            res: Residue, optional
                Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
            outname: string, optional
                Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_residues.png.
            camera: pyvista.Camera, optional
                Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to [0, 4 * "max distance from the center", 0] (see: https://pro-vis.readthedocs.io/en/latest/tutorial.html for more detail). Default: None.
        
        Returns: 
            Pyvista.Plotter window
                Window with interactive plot.
        """
        
        if not outname:
            outname = '_residues'
        self.plot(atoms=0, box=box, vw=0, bonds=0, residues=1, res=res, title="Residues", outname='_residues', camera=camera)
        
    def plot_vw(self, box=0, res=0, outname=0, camera=None):
        """
        Plot Van-der-Waals radius of atoms as spheres. Spheres have a wireframe style to be able to view inner structure as well.
        To plot Van-der-Waals radii as solid spheres use the manual_plot() member function.
        
        Parameters:
            box: bool, optional
                ptional - If True bounding box also visualized, default: 0.
            res: Residue, optional
                Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
            outname: string, optional
                Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_atoms.png.
            camera: pyvista.Camera, optional
                Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to [0, 4 * "max distance from the center", 0] (see: https://pro-vis.readthedocs.io/en/latest/tutorial.html for more detail). Default: None.
        
        Returns: 
            Pyvista.Plotter window
                Window with interactive plot.
        """
        if not outname:
            outname = '_vw'
        self.plot(atoms=1, box=box, vw=1, bonds=0, residues=0, res=res, title="Van-der-Waals", camera=camera)
        
    def plot_bonds(self, box=0, res=0, outname=0, colorful=False, camera=None):
        """
        Plot only the bonds. By default all bonds will be plotted uniformly. 
        
        If the difference in bond types is of interest set the "colorful" variable to True.
        Coloring:
        - Single bonds: white
        - Double bonds: blue
        - Triple bonds: green
        - Amide bonds: red
        - Aromatic bonds: purple
        - Undefined/Anything else: black
        
        Parameters:
            box: bool, optional
                If True bounding box also visualized, default: 0.
            res: Residue, optional
                Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
            outname: string, optional
                Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_atoms.png.
            colorful: bool, optional
                If True bonds will be plotted in a colorful manner. If False all bonds are white. Default: False
            camera: pyvista.Camera, optional
                Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to [0, 4 * "max distance from the center", 0] (see: https://pro-vis.readthedocs.io/en/latest/tutorial.html for more detail). Default: None.
        
        Returns: 
            Pyvista.Plotter window
                Window with interactive plot.
        """
        b = bool(colorful) * 1 + 1
        
        if not outname:
            outname = '_bonds'
        self.plot(atoms=0, box=box, vw=0, bonds=b, residues=0, res=res, outname=outname, title="Bonds", camera=camera)
        
        
    def plot_backbone(self, box=0, res=0, outname=0, camera=None):
        """
        Plots the backbone (roughly the amide bonds) of the protein.
        
        Parameters:
            box: bool, optional
                If True bounding box also visualized, default: 0.
            res: Residue, optional
                Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
            outname: string, optional
                Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_backbone.png.
            camera: pyvista.Camera, optional
                Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to [0, 4 * "max distance from the center", 0] (see: https://pro-vis.readthedocs.io/en/latest/tutorial.html for more detail). Default: None.
        
        Returns: 
            Pyvista.Plotter window
                Window with interactive plot.
        """
        
        if not outname:
            outname = '_backbone'
        self.plot(atoms=0, box=box, vw=0, bonds=0, residues=0, res=res, bb=1, outname=outname, title="Backbone", camera=camera)





# imports
from genericpath import sameopenfile
import os
from os.path import exists
import numpy as np
from numpy.lib.twodim_base import tri
import torch
import pyvista as pv
from pyvtk import PolyData, PointData, CellData, Scalars, Vectors, VtkData
import trimesh
from subprocess import PIPE, Popen
import open3d as o3d
import enum
from os.path import exists
import pickle

from provis.utils.surface_utils import get_surface, compute_normal, prepare_trimesh, fix_trimesh, find_nearest_atom
from provis.utils.surface_feat import compute_surface_features
from provis.src.processing.file_converter import FileConverter
from provis.src.processing.data_handler import DataHandler

class SurfaceHandler:
    """
    The 'brain' of provis, when it comes to handling surfaces. 
    
    This class loads information from a variety of files and creates meshes to be plotted. 
    Upper level classes - eg. Surface - use SurfaceHandler objects to create the meshes.
    
    The SurfaceHandler class loads atom-positional information from a .pdb file and from this information computes the surface mesh.
    The surface can be computed by the MSMS binary or natively. The MSMS version is chemically more accurate and faster, but the MSMS binary has to be downloaded for it to work.
    """
  
    def __init__(self, nc, fc=None, dh=None, density=3.0):
        """
        Initializes SurfaceHandler class memeber variables: 
        _dh: DataHandler - needed for native mesh creation.
        _fc: FileConverter - needed to create necessairy files.   
        _path: str - path to pdb file (without extension).
        _out_path: str - path to output files (without extension).
        _base_path: str - path to root directory.
        _density: float - density of surface mesh.
        _features: tuple(numpy.ndarray) - A collection of arrays representing the feature information for the surface of the protein, saved in a tuple.
        _mesh: Mesh - Surface mesh of protein.
        _col: numpy.ndarray - Coloring map of surface. To be passed to pyvista.PolyData.add_mesh(scalars=).
        _res_id: list - List of residue IDs (format from output_pdb_as_xyzrn())
        _atom_coords: list - List of atomic coordinates. Just the coordinates in the same order as in _res_id.

        Parameters:
            nc: NameChecker
                Instance of a NameChecker class. Used to pass the pdb file name and paths.
            fc: FileConverter
                Instance of a FileConverter class. Used to create the necessairy files (face, vert, pqr, ...) if they do not exist yet. Default: None.
            dh: DataHandler, optional
                Instance of a DataHandler class. Used to retrieve atom-positional information when calculating the surface of the protein natively. Default: None. If None a new DataHandler variable will be initialized with "nc".
            density: float, optional
                Density needed for msms. Default: 3.0.
        """
        
        if not fc:
            self._fc = FileConverter(nc)
        else:
            self._fc = fc
        if not dh:
            self._dh = DataHandler(nc, fc)
        else:
            self._dh = dh
        self._path, self._out_path, self._base_path, self._mesh_path = nc.return_all()
        self._density = density
        self._features = None
        self._mesh = None
        self._col = None
        self._res_id = None
        self._atom_coords = None
        
    def get_assignments(self):
        """
        Get assignments (coloring) for the mesh. File has to exist, no way to produce it with provis. 
        Loads "root directory"/data/tmp/{pdb_id}.pth and returns it.

        Returns: 
            PyTourch object
                Coloring of surface.
        """
        filename = self._out_path.upper() + '.pth'
        assigment = torch.load(filename)
        
        return assigment

    def get_surface_features(self, mesh, feature,  res_id=None):
        """
        Get the coloring corresponding to a specific feature.

        The code in words:
        Creates the pqr file if it does not exist.
        If the res_id variable is not empty retrieve the surface structure (list of specific surface-related information) using the get_surface() function.
        Else compile the surface structure from the mesh and the find_nearest_atom() function.
        Next, using the above mentioned surface structure compute all surface feature information.
        Finally return the coloring corresponding to the specified feature.

        Parameters:
            mesh: Trimesh
                The mesh
            feature: str
                Name of feature we are interested in. Options: hydrophob, shape, charge, hbonds.
            res_id: bool, optional
                List of unique residue IDs (format from output_pdb_as_xyzrn()). Default: False.
        
        Returns: 
            numpy.ndarray
                Array of coloring corresponding to surface.

        Raises:
            NotImplementedError
                If unkown feature specified error is raised
        """

        # get surface
        pdb_file_path = self._path 
        
        pqr_file_path = self._out_path
        if self._color_needed and self._dynamic:
            pdb_file_path = self._path + "_" + str(self._model_id) 
            pqr_file_path = self._out_path + "_" + str(self._model_id)
        if not exists(pqr_file_path + '.pqr'):
            self._fc.pdb_to_pqr(pdb_file_path, pqr_file_path)
        pdb_file = pdb_file_path + '.pdb'
        if not res_id:
            surface = get_surface(pqr_file_path, density=self._density)
        else:
            full_res_id = find_nearest_atom(self._atom_coords, res_id, mesh.vertices)
            
            surface = [mesh.vertices, mesh.faces, mesh.face_normals, full_res_id, mesh.area_faces]
            
        # compute features of surface
        if not self._features or self._color_needed:
            self._features = compute_surface_features(surface, pdb_file, pqr_file_path, mesh, pdb_id = pdb_file_path)

        if feature == 'hydrophob':
            return self._features[2]
        elif feature == 'shape':
            return self._features[0]
        elif feature == 'charge':
            return self._features[3]
        elif feature == 'hbonds':
            return self._features[1]
        else:
            raise NotImplementedError

    def msms_mesh_and_color(self, feature=None, patch=False):
        """
        Return the mesh and coloring created by the MSMS binary.

        The code in words:
        If self._mesh_needed is set to True - if the mesh could not be loaded from a file - compute the mesh using the .face and .vert files created by the MSMS binary.
        If self._dynamic is False - if the mesh is needed for a static molecule - and the .face and .vert files do not exist compute these files using the FileConverter class.
        Do the same for self._dynamic == True.
        If the mesh is needed compute the mesh from the .face and .vert files.
        Finally, if a feature is specified check if the color information could be loaded from a file (self._color_needed) and compute it if needed.
        If no feature specified set the variable that stores color information to None. This will result in a white mesh.

        Parameters:
            feature: str, optional
                Name of feature, same as in get_surface_features. Options: hydrophob, shape, charge, hbonds. Defaults to None.
            patch: bool, optional
                Set coloring of mesh manually, from a file. If set to True get_assignments() will be called. Defaults to False.
        """
        print("MSMS mesh calculation for model id: ", self._model_id)

        path = f"{self._out_path}_out_{int(self._density * 10)}"
        face = path + '.face'
        vert = path + '.vert'
        # Check if face file exists. If not convert it from pdb
        file_exists = exists(face) and exists(vert)
        if self._mesh_needed:
            if not self._dynamic:
                if not file_exists:
                    self._fc.msms(self._out_path, self._density)
            
            new_xyzrn_path = self._out_path
            if self._dynamic:
                new_xyzrn_path = self._out_path + "_" + str(self._model_id) 
                
                # convert current mesh
                path = f"{new_xyzrn_path}_out_{int(self._density * 10)}"
                face = path + '.face'
                vert = path + '.vert'
                # Check if face file exists. If not convert it from pdb
                file_exists = exists(face) and exists(vert)
                if not file_exists:
                    print("Creating face and vert files for current model...")
                    new_name = self._path + "_" + str(self._model_id) 
                    
                    self._fc.pdb_to_xyzrn(new_name, new_xyzrn_path)
                    self._fc.msms(new_xyzrn_path, self._density)
                    path = f"{new_xyzrn_path}_out_{int(self._density * 10)}"
                
            print(" - Get surface")
            surface = get_surface(out_path=new_xyzrn_path, density=self._density)
            self._mesh = prepare_trimesh(vertices=surface[0], faces=surface[1], normals=surface[2], 
                            resolution=1.5, apply_fixes=True)
        
            meshname = self._mesh_path + "_msms_" + str(self._model_id) + '.obj'
            self._mesh.export(meshname)
        
        if feature:
            if self._color_needed:
                print(" - Feature: ", feature)
                if patch:
                    self._col = self.get_assignments()
                if not patch:
                    self._col = self.get_surface_features(self._mesh, feature)
                fname = self._mesh_path + "_msms_" + feature + "_" + str(self._model_id)
                with open(fname, 'wb') as f:
                    pickle.dump(self._col, f)
        else:
            self._col = None

    def native_mesh_and_color(self, feature=None):
        """
        Returns a mesh without the need for the MSMS binary. The mesh and coloring is also saved to the following file names:
        Mesh: "root directory"/data/meshes/{pdb_id}_{model_id}.obj
        Color: "root directory"/data/meshes/{pdb_id}_{feature}_{model_id}
        Always returns a pyvista.PolyData mesh of the surface and if a feature is specified it also returns the coloring according to that feature.

        Parameters:
            feature: str, optional
                Name of feature, same as in get_surface_features. Options: hydrophob, shape, charge, hbonds. Defaults to "".
        """
        print("Native mesh calculation for model id: ", self._model_id)
        if self._mesh_needed:

            print(" - Feature: ", feature)
            # create rough surface by combining vw radii of each atom
            atom_data, self._res_id, self._atom_coords = self._dh.get_atoms_IDs(model_id=self._model_id)
            self._atmsurf, col, _ = self._dh.get_atom_mesh(atom_data, vw=1, probe=0.1)

            # adding the spheres (by atom type) one at a time
            j = 0
            mesh_ = pv.wrap(self._atmsurf[0])
            for mesh in self._atmsurf[1:]:
                mesh_ = mesh_ + (mesh)
            
            # blur the spheres and extract edges (pyvista)
            mesh_ = mesh_.extract_surface()#.extract_feature_edges(1)#.delaunay_3d(alpha=a).extract_feature_edges(2)
            # create trimesh mesh
            new = trimesh.Trimesh(mesh_.points)
            cloud = o3d.geometry.PointCloud()
            cloud.points = o3d.utility.Vector3dVector(new.vertices)
            # cloud.normals = o3d.utility.Vector3dVector(new.vertex_normals)
            cloud.estimate_normals()

            # to obtain a consistent normal orientation
            cloud.orient_normals_towards_camera_location(cloud.get_center())

            # or you might want to flip the normals to make them point outward, not mandatory
            cloud.normals = o3d.utility.Vector3dVector( - np.asarray(cloud.normals))
            # reconstruct the surface mesh as a trimesh mesh
            tri_mesh, _ = o3d.geometry.TriangleMesh.create_from_point_cloud_poisson(cloud, depth=5)
            print(" - Poisson completed")
            v = np.asarray(tri_mesh.vertices)
            f = np.array(tri_mesh.triangles)

            tri_mesh = trimesh.Trimesh(v, faces=f)

            self._mesh = tri_mesh
            meshname = self._mesh_path + "_" + str(self._model_id) + '.obj'
            tri_mesh.export(meshname)
            print("Mesh calculation done.")
            

        # only needed if feature is specified (-> color map needs to be calculated)
        if feature:
            if self._color_needed:
                self._col = self.get_surface_features(self._mesh, feature, self._res_id)
                
                fname = self._mesh_path + "_" + feature + "_" + str(self._model_id)
                with open(fname, 'wb') as f:
                    pickle.dump(self._col, f)
        else:
            self._col = None

    
    def return_mesh_and_color(self, msms=False, feature=None, patch=False, model_id=0, num_models=0):
        """
        Wrapper function to choose between the msms surface visualization vs the native surface visualization. 
        If you could not download the MSMS binary leave the msms variable as False.
        
        If the mesh and appropriate coloring has already been stored to a file, then this information will be loaded and no computation will be done.
        
        The code in words:
        First, set a few class member variables that are used in the {*}_mesh_and_color() member functions.
        Next check if the mesh and color information has already been computed. If the files already exist load them and return. No computation will be done.
        Otherwise, depending on what the msms input variable is set to, compute either the msms_mesh_and_color() ir the native_mesh_and_color().
        Return the mesh and color information.

        Parameters:
            msms: bool, optional
                If true, surface generated by msms binary is returned, else the native mesh. Default: False.
            feature: str, optional
                Name of feature, same as in get_surface_features. Options: hydrophob, shape, charge, hbonds. Defaults to None.
            patch: bool, optional
                Set coloring of mesh manually. If set to True get_assignments() will be called. Defaults to False.
            model_id: int, optional
                The dynamic model ID of the desired molecule. Count starts at 0. Leave default value for static molecules. Default: 0.
            num_models: int, optional
                Number of models in a trajectory. Only important for dynamic structures. Default: 0.
        
        Returns: 
            trimesh.Trimesh
                The mesh corresponding to the surface of the protein.
            numpy.ndarray
                Coloring map corresponding to specified feature.
        """
        self._color_needed = True
        self._mesh_needed = True
        self._msms = msms
        self._model_id = model_id
        self._dynamic = num_models > 0
        connect = "_"
        if msms:
            connect = "_msms_"
            
        meshname = self._mesh_path + connect + str(model_id) + '.obj'
        mesh_exists = exists(meshname)
        if feature:
            fname = self._mesh_path + connect + feature + "_" + str(model_id)
            col_exists = exists(fname)
            if mesh_exists and col_exists:
                print("Loading mesh from: ", meshname)
                print("Loading features from: ", fname)
                self._mesh = trimesh.load_mesh(meshname)
                self._mesh_needed = False
                with open(fname, 'rb') as f:
                    self._col = pickle.load(f)
                self._color_needed = False
        else:
            if mesh_exists:
                self._mesh = trimesh.load_mesh(meshname)
                print("Loading mesh from: ", meshname)
                self._mesh_needed = False
           
        if self._color_needed or self._mesh_needed:
            # run appropriate function to compute the mesh and the coloring
            if msms:
                self.msms_mesh_and_color(feature, patch)
            else:
                self.native_mesh_and_color(feature)

        return self._mesh, self._col



"""
https://github.com/bunnech/holoprot/blob/main/holoprot/feat/surface.py is the base for this file. Modifications were made.

Functions to compute features for a patch on the protein surface.

Some of these are borrowed from https://github.com/LPDI-EPFL/masif under the https://github.com/LPDI-EPFL/masif/blob/master/LICENSE license
"""
import numpy as np
from Bio.PDB import Selection
from Bio.PDB.Residue import Residue
from subprocess import PIPE, Popen
import os
from scipy.spatial import KDTree
from typing import List, Tuple, Dict
import trimesh

from provis.src.processing import KD_SCALE, get_residues
from provis.utils.charges_utils import (compute_hbond_helper, compute_satisfied_CO_HN,
                    normalize_electrostatics)
from provis.utils.surface_utils import prepare_trimesh


Surface = Tuple[np.ndarray, np.ndarray, np.ndarray, List[str], Dict[str, str]]

def compute_shape_index(mesh) -> np.ndarray:
    """
    Computes shape index for the patches. Shape index characterizes the shape
    around a point on the surface, computed using the local curvature around each
    point. These values are derived using Trimesh's available geometric
    processing functionality.

    Parameters:
        mesh: Trimesh
            The mesh is constructed using information about vertices and faces.

    Returns: 
        np.ndarray
            Shape index for each vertex
    """
    normals = mesh.vertex_normals
    # n1 = mesh.get_attribute("vertex_nx")
    # n2 = mesh.get_attribute("vertex_ny")
    # n3 = mesh.get_attribute("vertex_nz")
    # normals = np.stack([n1, n2, n3], axis=1)
    # mesh.add_attribute("vertex_mean_curvature")
    # H = mesh.get_attribute("vertex_mean_curvature")
    # mesh.add_attribute("vertex_gaussian_curvature")
    # K = mesh.get_attribute("vertex_gaussian_curvature")
    H = trimesh.curvature.discrete_mean_curvature_measure(mesh, mesh.vertices, 1)
    K = trimesh.curvature.discrete_gaussian_curvature_measure(mesh, mesh.vertices, 1)
    elem = np.square(H) - K
    # In some cases this equation is less than zero, likely due to the method
    # that computes the mean and gaussian curvature. set to an epsilon.
    elem[elem < 0] = 1e-8
    k1 = H + np.sqrt(elem)
    k2 = H - np.sqrt(elem)
    # Compute the shape index
    si = (k1 + k2) / (k1 - k2)
    si = np.arctan(si) * (2 / np.pi)
    return si


def compute_hbonds(vertices: np.ndarray, residues: List[Residue],
                   names: List[str]) -> np.ndarray:
    """Compute H-bond (hydrogen-bond) induced charges at every vertex.
    
    Parameters:
        vertices: np.ndarray
            Vertices of mesh
        residues: List[Residue]
            List of residues to compute
        names: List[str]
            List of custom names created by output_pdb_as_xyzrn()
    
    Returns: 
        np.ndarray
            Array of bonds, by vertex
    """
    
    residue_dict = {}
    for res in residues:
        chain_id = res.get_parent().get_id()
        if chain_id == "":
            chain_id = " "
        residue_dict[(chain_id, res.get_id())] = res

    atoms = Selection.unfold_entities(residues, "A")
    satisfied_CO, satisfied_HN = compute_satisfied_CO_HN(atoms)

    hbond = np.array([0.0] * len(vertices))
    # Go over every vertex
    for ix, name in enumerate(names):
        fields = name.split("_")
        chain_id = fields[0]
        if chain_id == "":
            chain_id = " "
        if fields[2] == "x":
            fields[2] = " "
        res_id = (" ", int(fields[1]), fields[2])
        aa = fields[3]
        atom_name = fields[4]
        # Ignore atom if it is BB and it is already satisfied.
        if atom_name == "H" and res_id in satisfied_HN:
            continue
        if atom_name == "O" and res_id in satisfied_CO:
            continue
        # Compute the charge of the vertex
        hbond[ix] = compute_hbond_helper(
            atom_name, residue_dict[(chain_id, res_id)], vertices[ix])

    return hbond


def assign_props_to_new_mesh(new_vertices,
                             old_vertices: np.ndarray,
                             old_props: np.ndarray,
                             feature_interpolation: bool = True) -> np.ndarray:
    """
    Assign properties to vertices in modified mesh given the initial mesh. The
    assignment is carried using a KDTree data structure to query nearest points.

    Parameters:
        new_vertices: np.ndarray 
            Vertices on the modified mesh
        old_vertices: np.ndarray 
            Vertices on the original mesh
        old_props: np.ndarray 
            Property values for each vertex on the original mesh
        feature_interpolation: bool, (default True)
            If set to True interpolates features to new vertices.

    Returns: 
        np.ndarray
            Property values for vertices on the modified mesh
    """
    dataset = old_vertices
    testset = new_vertices
    new_props = np.zeros(len(new_vertices))
    if feature_interpolation:
        num_inter = 4  # Number of interpolation features
        # Assign k old vertices to each new vertex.
        kdt = KDTree(dataset)
        dists, result = kdt.query(testset, k=num_inter)
        # Square the distances (as in the original pyflann)
        dists = np.square(dists)
        # The size of result is the same as new_vertices
        for vi_new in range(len(result)):
            vi_old = result[vi_new]
            dist_old = dists[vi_new]
            # If one vertex is right on top, ignore the rest.
            if dist_old[0] == 0.0:
                new_props[vi_new] = old_props[vi_old[0]]
                continue

            total_dist = np.sum(1 / dist_old)
            for i in range(num_inter):
                new_props[vi_new] += (
                    old_props[vi_old[i]] * (1 / dist_old[i]) / total_dist)
    else:
        # Assign k old vertices to each new vertex.
        kdt = KDTree(dataset)
        dists, result = kdt.query(testset)
        new_props = old_props[result]
    return new_props


def compute_hydrophobicity(names: List[str]) -> np.ndarray:
    """
    Compute hydrophobicity value for all vertices on the surface. Each surface
    vertex has a mapping to the corresponding residue from the original protein.
    This is used to assign a hydrophobicity value to each vertex using the Kyte-
    Doolittle scale.

    Parameters:
        names: List[str]
            Identifier names for each vertex in the surface
    
    Returns: 
        np.ndarray
            Hydrophobicity values for each surface vertex
    """
    hp = np.zeros(len(names))
    for ix, name in enumerate(names):
        res_name = name.split("_")[3]
        hp[ix] = KD_SCALE[res_name] / 4.5
    return hp


def compute_charges(vertices: np.ndarray, pdb_id: str, path: str) -> np.ndarray:
    """
    Computes electrostatics for the surface vertices. The function first calls
    the PDB2PQR executable to prepare the pdb file for electrostatics. Poisson-
    Boltzmann electrostatics are computed using APSB executable. Multivalue,
    provided within APSB suite is used to assign charges to each vertex. The
    charges are further normalized.

    Parameters:
        vertices: np.ndarray
            Surface vertex coordinates
        pdb_id: str
            PDB ID of the protein
        path: str
            Path of the pqr file in the form {path}.pqr
    
    Returns: 
        np.ndarray
            Charge values for each vertex
    """

    charge_file = f"{path}.pqr"
    if not os.path.exists(charge_file):
        raise ValueError(f"Charges cannot be computed. Missing file {pdb_id}.pqr. {pdb_id}")
        return None

    chargefile = open(charge_file)
    charges = np.array([0.0] * len(vertices))
    for ix, line in enumerate(chargefile.readlines()):
        try:
            charges[ix] = float(line.split()[8]) # "," for some reson pqr does not have commas, but whitespaces
        except:
            pass
    charges = normalize_electrostatics(charges)
    return charges


def compute_surface_features(surface: Surface,
                             pdb_file: str,
                             path: str,
                             mesh = None,
                             fix_mesh: bool = False,
                             return_mesh: bool = False,
                             pdb_id: str = None) -> Tuple[np.ndarray]:
    """Computes all patch features.

    Parameters:
        surface: Surface
            Tuple of attributes characterizing the surface. These include vertices,
            faces, normals to each vertex, areas, fdue identifiers for vertices.
        pdb_file: str
            PDB File containing the atomic coordinates
        path: str
            Path to files without extensions. Usually data/tmp/{pdb_id}
        mesh: trimesh.Trimesh
            The mesh.
        fix_mesh: bool, optional
            Whether to fix the mesh by collapsing nodes and edges. Default: False.
        return_mesh: bool, optional
            Whether to return the mesh. Default: False.
        pdb_id: str, optional
            PDB id of the associated protein. Default: None

    Returns: 
        np.ndarray
            Shape index
        np.ndarray
            Hydrogen bond induced charges
        np.ndarray
            Hydrophobicity of each surface vertex
        np.ndarray
            Electrostatics of each surface vertex
    """
    residues = get_residues(pdb_file)
    vertices, faces, normals, names, areas = surface
    hbonds = compute_hbonds(vertices, residues, names)
    hydrophob = compute_hydrophobicity(names)
    if mesh is None:
        mesh = prepare_trimesh(vertices, faces, normals=normals, apply_fixes=fix_mesh)
    si = compute_shape_index(mesh)
    
    if fix_mesh or (len(mesh.vertices) != len(vertices)):
        hbonds = assign_props_to_new_mesh(
            new_vertices=mesh.vertices,
            old_vertices=vertices,
            old_props=hbonds,
            feature_interpolation=True)
        hydrophob = assign_props_to_new_mesh(
            new_vertices=mesh.vertices,
            old_vertices=vertices,
            old_props=hydrophob,
            feature_interpolation=True)

    charges = compute_charges(mesh.vertices, pdb_id, path)

    if si is None or not np.isfinite(si).all():
        raise ValueError(f"{pdb_id}: Shape index failed nan check.")

    if hbonds is None or not np.isfinite(hbonds).all():
        raise ValueError(f"{pdb_id}: HBonds failed nan check.")

    if hydrophob is None or not np.isfinite(hydrophob).all():
        raise ValueError(f"{pdb_id}: Hydrophob failed nan check.")

    if charges is None or not np.isfinite(charges).all():
        raise ValueError(f"{pdb_id}: Charges failed nan check.")

    node_feats = (si, hbonds, hydrophob, charges)
    
    if return_mesh:
        return node_feats, mesh

    return node_feats
